#! /usr/bin/env python

"""
Process the examples in the documentation for inclusion in the Gallery:

- create a "documentation" directory within "examples"
- add a README.txt file
- copy the examples from the documentation, removing the "doc_" from the
   filename
- add the required docstring to the files for proper rendering
- copy the data files

"""

import os
from pathlib import Path
import shlex
from shutil import copy2
import subprocess


def copy_data_files(src_dir, dest_dir):
    """Copy files with datafile extension from src_dir to dest_dir."""
    data_file_extension = [".dat", ".csv", ".sav"]

    for file in src_dir.glob("*"):
        if file.suffix in data_file_extension:
            copy2(file, dest_dir)


doc_dir = Path(__file__).parent.absolute()

examples_dir = doc_dir.parent / "examples"
files = examples_dir.glob("doc[_]*.py")

examples_documentation_dir = examples_dir / "documentation"
examples_documentation_dir.mkdir(exist_ok=True)


scripts_to_run = []

(examples_documentation_dir / "README.txt").write_text(
    "Examples from the documentation\n"
    "===============================\n\n"
    "Below are all the examples that are part of the lmfit documentation."
)

for fn in files:
    sname = fn.name[4:]
    lmfit_class, *description = sname[:-3].split('_')
    gallery_name = f"{lmfit_class.capitalize()} - {' '.join(description)}"

    script_text = fn.read_text()

    gallery_file = examples_documentation_dir / sname
    msg = ""  # add optional message
    gallery_file.write_text(f'"""\n{gallery_name}\n{"=" * len(gallery_name)}\n\n'
                            f'{msg}\n"""\n{script_text}')

    # make sure the saved Models and ModelResult are available
    if "save" in fn.name:
        scripts_to_run.append(gallery_file)

copy_data_files(examples_dir, examples_documentation_dir)

os.chdir(examples_documentation_dir)

for script in scripts_to_run:
    subprocess.run(shlex.split(f"python {script.as_posix()}"), check=True)

os.chdir(doc_dir)

# data files for the other Gallery examples
copy_data_files(examples_documentation_dir, doc_dir)



# <examples/doc_fitting_withreport.py>
from numpy import exp, linspace, pi, random, sign, sin

from lmfit import create_params, fit_report, minimize

p_true = create_params(amp=14.0, period=5.46, shift=0.123, decay=0.032)


def residual(pars, x, data=None):
    """Model a decaying sine wave and subtract data."""
    vals = pars.valuesdict()
    amp = vals['amp']
    per = vals['period']
    shift = vals['shift']
    decay = vals['decay']

    if abs(shift) > pi/2:
        shift = shift - sign(shift)*pi
    model = amp * sin(shift + x/per) * exp(-x*x*decay*decay)
    if data is None:
        return model
    return model - data


random.seed(0)
x = linspace(0.0, 250., 1001)
noise = random.normal(scale=0.7215, size=x.size)
data = residual(p_true, x) + noise

fit_params = create_params(amp=13, period=2, shift=0, decay=0.02)

out = minimize(residual, fit_params, args=(x,), kws={'data': data})

print(fit_report(out))
# <end examples/doc_fitting_withreport.py>



"""Tests for the basinhopping minimization algorithm."""
import numpy as np
from numpy.testing import assert_allclose
import pytest
from scipy.optimize import basinhopping

import lmfit


def test_basinhopping_lmfit_vs_scipy():
    """Test basinhopping in lmfit versus scipy."""
    # SciPy
    def func(x):
        return np.cos(14.5*x - 0.3) + (x+0.2) * x

    minimizer_kwargs = {'method': 'L-BFGS-B'}
    x0 = [1.]

    ret = basinhopping(func, x0, minimizer_kwargs=minimizer_kwargs, seed=7)

    # lmfit
    def residual(params):
        x = params['x'].value
        return np.cos(14.5*x - 0.3) + (x+0.2) * x

    pars = lmfit.Parameters()
    pars.add_many(('x', 1.))
    kws = {'minimizer_kwargs': {'method': 'L-BFGS-B'}, 'seed': 7}
    mini = lmfit.Minimizer(residual, pars)
    out = mini.minimize(method='basinhopping', **kws)

    assert_allclose(out.residual, ret.fun)
    assert_allclose(out.params['x'].value, ret.x, rtol=1e-5)


def test_basinhopping_2d_lmfit_vs_scipy():
    """Test basinhopping in lmfit versus scipy."""
    # SciPy
    def func2d(x):
        return np.cos(14.5*x[0] - 0.3) + (x[1]+0.2) * x[1] + (x[0]+0.2) * x[0]

    minimizer_kwargs = {'method': 'L-BFGS-B'}
    x0 = [1.0, 1.0]

    ret = basinhopping(func2d, x0, minimizer_kwargs=minimizer_kwargs, seed=7)

    # lmfit
    def residual_2d(params):
        x0 = params['x0'].value
        x1 = params['x1'].value
        return np.cos(14.5*x0 - 0.3) + (x1+0.2) * x1 + (x0+0.2) * x0

    pars = lmfit.Parameters()
    pars.add_many(('x0', 1.), ('x1', 1.))

    mini = lmfit.Minimizer(residual_2d, pars)
    kws = {'minimizer_kwargs': {'method': 'L-BFGS-B'}, 'seed': 7}
    out = mini.minimize(method='basinhopping', **kws)

    assert_allclose(out.residual, ret.fun)
    assert_allclose(out.params['x0'].value, ret.x[0], rtol=1e-5)
    assert_allclose(out.params['x1'].value, ret.x[1], rtol=1e-5)


def test_basinhopping_Alpine02(minimizer_Alpine02):
    """Test basinhopping on Alpine02 function."""
    global_optimum = [7.91705268, 4.81584232]
    fglob = -6.12950

    kws = {'minimizer_kwargs': {'method': 'L-BFGS-B'}, 'seed': 7}
    out = minimizer_Alpine02.minimize(method='basinhopping', **kws)
    out_x = np.array([out.params['x0'].value, out.params['x1'].value])

    assert_allclose(out.residual, fglob, rtol=1e-5)
    assert_allclose(min(out_x), min(global_optimum), rtol=1e-3)
    assert_allclose(max(out_x), max(global_optimum), rtol=1e-3)
    assert out.method == 'basinhopping'


def test_basinhopping_bounds(minimizer_Alpine02):
    """Test basinhopping algorithm with bounds."""
    # change boundaries of parameters
    pars_bounds = lmfit.Parameters()
    pars_bounds.add_many(('x0', 1., True, 5.0, 15.0),
                         ('x1', 1., True, 2.5, 7.5))

    kws = {'minimizer_kwargs': {'method': 'L-BFGS-B'}, 'seed': 7}
    out = minimizer_Alpine02.minimize(params=pars_bounds,
                                      method='basinhopping', **kws)
    assert 5.0 <= out.params['x0'].value <= 15.0
    assert 2.5 <= out.params['x1'].value <= 7.5


def test_basinhopping_solver_options(minimizer_Alpine02):
    """Test basinhopping algorithm, pass incorrect options to solver."""
    # use minimizer_kwargs to pass an invalid method for local solver to
    # scipy.basinhopping
    kws = {'minimizer_kwargs': {'method': 'unknown'}}
    with pytest.raises(ValueError, match=r'Unknown solver'):
        minimizer_Alpine02.minimize(method='basinhopping', **kws)

    # pass an incorrect value for niter to scipy.basinhopping
    kws = {'niter': 'string'}
    with pytest.raises(TypeError):
        minimizer_Alpine02.minimize(method='basinhopping', **kws)



import os

import numpy as np
import pytest

import lmfit


@pytest.fixture
def minimizer_Alpine02():
    """Return a lmfit Minimizer object for the Alpine02 function."""
    def residual_Alpine02(params):
        x0 = params['x0'].value
        x1 = params['x1'].value
        return np.prod(np.sqrt(x0) * np.sin(x0)) * np.prod(np.sqrt(x1) *
                                                           np.sin(x1))

    # create Parameters and set initial values and bounds
    pars = lmfit.Parameters()
    pars.add_many(('x0', 1., True, 0.0, 10.0),
                  ('x1', 1., True, 0.0, 10.0))

    mini = lmfit.Minimizer(residual_Alpine02, pars)
    return mini


@pytest.fixture
def peakdata():
    """Return the peak-like test data."""
    data = np.loadtxt(os.path.join(os.path.dirname(__file__), '..',
                                   'examples', 'test_peak.dat'))
    return data.T



# <examples/doc_model_loadmodelresult.py>
import os
import sys

import matplotlib.pyplot as plt
import numpy as np

from lmfit.model import load_modelresult

if not os.path.exists('gauss_modelresult.sav'):
    os.system(f"{sys.executable} doc_model_savemodelresult.py")

data = np.loadtxt('model1d_gauss.dat')
x = data[:, 0]
y = data[:, 1]

result = load_modelresult('gauss_modelresult.sav')
print(result.fit_report())

plt.plot(x, y, 'o')
plt.plot(x, result.best_fit, '-')
plt.show()
# <end examples/doc_model_loadmodelresult.py>



"""
Benchmarks of methods with and without computing the Jacobian analytically
==========================================================================

Providing a function that calculates the Jacobian matrix analytically can
reduce the time spent finding a solution. The results from benchmarks comparing
two methods (``leastsq`` and ``least_squares``) with and without a function to
calculate the Jacobian matrix analytically are presented below.

First we define the model function, the residual function, and the appropriate
Jacobian functions:
"""
from timeit import timeit
from types import SimpleNamespace

import matplotlib.pyplot as plt
import numpy as np

from lmfit import Parameters, minimize

NUM_JACOBIAN_CALLS = 0


def func(var, x):
    return var[0] * np.exp(-var[1]*x) + var[2]


def residual(pars, x, data):
    a, b, c = pars['a'], pars['b'], pars['c']
    model = func((a, b, c), x)
    return model - data


def dfunc(pars, x, data):
    global NUM_JACOBIAN_CALLS
    NUM_JACOBIAN_CALLS += 1

    a, b = pars['a'], pars['b']
    v = np.exp(-b*x)
    return np.array([v, -a*x*v, np.ones(len(x))])


def jacfunc(pars, x, data):
    global NUM_JACOBIAN_CALLS
    NUM_JACOBIAN_CALLS += 1

    a, b = pars['a'], pars['b']
    v = np.exp(-b*x)
    jac = np.ones((len(x), 3), dtype=np.float64)
    jac[:, 0] = v
    jac[:, 1] = -a * x * v
    return jac


a, b, c = 2.5, 1.3, 0.8

x = np.linspace(0, 4, 50)
y = func([a, b, c], x)

data = y + 0.15*np.random.RandomState(seed=2021).normal(size=x.size)


###############################################################################
# Then we define the different cases to benchmark (i.e., different methods with
# and without a function to calculate the Jacobian analytically) and the number
# of repetitions per case:
cases = (
    dict(
        method='leastsq',
    ),
    dict(
        method='leastsq',
        Dfun=dfunc,
        col_deriv=1,
    ),
    dict(
        method='least_squares',
    ),
    dict(
        method='least_squares',
        jac=jacfunc,
    ),
)

num_repeats = 100
results = []

for kwargs in cases:
    params = Parameters()
    params.add('a', value=10)
    params.add('b', value=10)
    params.add('c', value=10)

    wrapper = lambda: minimize(
        residual,
        params,
        args=(x,),
        kws={'data': data},
        **kwargs,
    )
    time = timeit(wrapper, number=num_repeats) / num_repeats

    NUM_JACOBIAN_CALLS = 0
    fit = wrapper()

    results.append(SimpleNamespace(
        time=time,
        num_jacobian_calls=NUM_JACOBIAN_CALLS,
        fit=fit,
        kwargs=kwargs,
    ))


###############################################################################
# Finally, we present the results:
labels = []

for result in results:
    label = result.kwargs['method']
    if result.num_jacobian_calls > 0:
        label += ' with Jac.'

    labels.append(label)

label_width = max(map(len, labels))
lines = [
    '| '
    + ' | '.join([
        'Method'.ljust(label_width),
        'Avg. time (ms)',
        '# func. (+ Jac.) calls',
        'Chi-squared',
        'a'.ljust(5),
        'b'.ljust(5),
        'c'.ljust(6),
    ])
    + '|'
]

print(f'The "true" parameters are: a = {a:.3f}, b = {b:.3f}, c = {c:.3f}\n')
fig, ax = plt.subplots()
ax.plot(x, data, marker='.', linestyle='none', label='data')

for (result, label) in zip(results, labels):
    linestyle = '-'
    if result.num_jacobian_calls > 0:
        linestyle = '--'

    a = result.fit.params['a'].value
    b = result.fit.params['b'].value
    c = result.fit.params['c'].value
    y = func([a, b, c], x)
    ax.plot(x, y, label=label, alpha=0.5, linestyle=linestyle)

    columns = [
        label.ljust(label_width),
        f'{result.time * 1000:.2f}'.ljust(14),
        (
            f'{result.fit.nfev}'
            + (
                f' (+{result.num_jacobian_calls})'
                if result.num_jacobian_calls > 0 else
                ''
            )
        ).ljust(22),
        f'{result.fit.chisqr:.3f}'.ljust(11),
        f'{a:.3f}'.ljust(5, '0'),
        f'{b:.3f}'.ljust(5, '0'),
        f'{c:.3f}'.ljust(5, '0'),
    ]
    lines.append('| ' + ' | '.join(columns) + ' |')

lines.insert(1, '|-' + '-|-'.join('-' * len(col) for col in columns) + '-|')
print('\n'.join(lines))

ax.set_xlabel('x')
ax.set_ylabel('y')
ax.legend()



"""
Outlier detection via leave-one-out
===================================

Outliers can sometimes be identified by assessing the influence of each
datapoint. To assess the influence of one point, we fit the dataset without the
point and compare the result with the fit of the full dataset. The code below
shows how to do this with lmfit. Note that the presented method is very basic.
"""
from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np

import lmfit

###############################################################################
# Generate test data and model:
x = np.linspace(0.3, 10, 100)
np.random.seed(1)
y = 1.0 / (0.1 * x) + 2.0 + 3 * np.random.randn(x.size)

params = lmfit.Parameters()
params.add_many(('a', 0.1), ('b', 1))


def func(x, a, b):
    return 1.0 / (a * x) + b


###############################################################################
# Make five points outliers:
idx = np.random.randint(0, x.size, 5)
y[idx] += 10 * np.random.randn(idx.size)

###############################################################################
# Fit the data:
model = lmfit.Model(func, independent_vars=['x'])
fit_result = model.fit(y, x=x, a=0.1, b=2)

###############################################################################
# and gives the plot and fitting results below:
fit_result.plot_fit()
plt.plot(x[idx], y[idx], 'o', label='outliers')
plt.show()

###############################################################################
print(fit_result.fit_report())

###############################################################################
# Fit the dataset while omitting one data point:
best_vals = defaultdict(lambda: np.zeros(x.size))
stderrs = defaultdict(lambda: np.zeros(x.size))
chi_sq = np.zeros_like(x)
for i in range(x.size):
    idx2 = np.arange(0, x.size)
    idx2 = np.delete(idx2, i)
    tmp_x = x[idx2]
    tmp = model.fit(y[idx2], x=tmp_x, a=fit_result.params['a'],
                    b=fit_result.params['b'])
    chi_sq[i] = tmp.chisqr
    for p in tmp.params:
        tpar = tmp.params[p]
        best_vals[p][i] = tpar.value
        stderrs[p][i] = (tpar.stderr / fit_result.params[p].stderr)

###############################################################################
# Plot the influence on the red. chisqr of each point:
fig, ax = plt.subplots()
ax.plot(x, (fit_result.chisqr - chi_sq) / chi_sq)
ax.scatter(x[idx], fit_result.chisqr / chi_sq[idx] - 1, color='r',
           label='outlier')
ax.set_ylabel(r'Relative red. $\chi^2$ change')
ax.set_xlabel('x')
ax.legend()

###############################################################################
# Plot the influence on the parameter value and error of each point:
fig, axs = plt.subplots(4, figsize=(4, 7), sharex='col')
axs[0].plot(x, best_vals['a'])
axs[0].scatter(x[idx], best_vals['a'][idx], color='r', label='outlier')
axs[0].set_ylabel('best a')

axs[1].plot(x, best_vals['b'])
axs[1].scatter(x[idx], best_vals['b'][idx], color='r', label='outlier')
axs[1].set_ylabel('best b')

axs[2].plot(x, stderrs['a'])
axs[2].scatter(x[idx], stderrs['a'][idx], color='r', label='outlier')
axs[2].set_ylabel('err a change')

axs[3].plot(x, stderrs['b'])
axs[3].scatter(x[idx], stderrs['b'][idx], color='r', label='outlier')
axs[3].set_ylabel('err b change')

axs[3].set_xlabel('x')



"""Tests for the calculation of confidence intervals."""
import copy

import numpy as np
from numpy.testing import assert_allclose
import pytest
from scipy.stats import f

import lmfit


@pytest.fixture
def data():
    """Generate synthetic data."""
    x = np.linspace(0.3, 10, 100)
    np.random.seed(0)
    y = 1.0 / (0.1 * x) + 2.0 + 0.1 * np.random.randn(x.size)
    return (x, y)


@pytest.fixture
def pars():
    """Create and initialize parameter set."""
    parameters = lmfit.Parameters()
    parameters.add_many(('a', 0.1), ('b', 1))
    return parameters


def residual(params, x, data):
    """Define objective function for the minimization."""
    model = 1.0 / (params['a'] * x) + params['b']
    return data - model


def test_default_f_compare(data, pars):
    """Test the default f_compare function: F-test."""
    minimizer = lmfit.Minimizer(residual, pars, fcn_args=(data))
    out = minimizer.leastsq()

    # "fixing" a parameter, keeping the chisqr the same
    out2 = copy.deepcopy(out)
    out2.nvarys = 1
    prob = lmfit.confidence.f_compare(out, out2)
    assert_allclose(prob, 0.0)

    # "fixing" a parameter, increasing the chisqr
    out2.chisqr = 1.0015*out.chisqr
    prob = lmfit.confidence.f_compare(out, out2)
    assert_allclose(prob, 0.2977506)


def test_copy_and_restore_vals(data, pars):
    """Test functions to save and restore parameter values and stderrs."""
    # test copy_vals without/with stderr present
    copy_pars = lmfit.confidence.copy_vals(pars)

    assert isinstance(copy_pars, dict)
    for _, par in enumerate(pars):
        assert_allclose(pars[par].value, copy_pars[par][0])
        assert copy_pars[par][1] is None  # no stderr present

    minimizer = lmfit.Minimizer(residual, pars, fcn_args=(data))
    out = minimizer.leastsq()
    copy_pars_out = lmfit.confidence.copy_vals(out.params)

    assert isinstance(copy_pars_out, dict)
    for _, par in enumerate(out.params):
        assert_allclose(out.params[par].value, copy_pars_out[par][0])
        assert_allclose(out.params[par].stderr, copy_pars_out[par][1])

    # test restore_vals to the original parameter set after changing them first
    pars['a'].set(value=1.0)
    pars['b'].set(value=10)
    lmfit.confidence.restore_vals(copy_pars, pars)

    assert isinstance(pars, lmfit.parameter.Parameters)
    assert_allclose(pars['a'].value, 0.1)
    assert_allclose(pars['b'].value, 1.0)
    assert pars['a'].stderr is None
    assert pars['b'].stderr is None

    lmfit.confidence.restore_vals(copy_pars_out, pars)
    for _, par in enumerate(pars):
        assert_allclose(pars[par].value, out.params[par].value)
        assert_allclose(pars[par].stderr, out.params[par].stderr)


@pytest.mark.parametrize("verbose", [False, True])
def test_confidence_leastsq(data, pars, verbose, capsys):
    """Calculate confidence interval after leastsq minimization."""
    minimizer = lmfit.Minimizer(residual, pars, fcn_args=(data))
    out = minimizer.leastsq()

    assert 5 < out.nfev < 500
    assert out.chisqr < 3.0
    assert out.nvarys == 2
    assert_allclose(out.params['a'], 0.1, rtol=0.01)
    assert_allclose(out.params['b'], 2.0, rtol=0.01)

    ci = lmfit.conf_interval(minimizer, out, verbose=verbose)
    assert_allclose(ci['b'][0][0], 0.997, rtol=0.01)
    assert_allclose(ci['b'][0][1], 1.947, rtol=0.01)
    assert_allclose(ci['b'][2][0], 0.683, rtol=0.01)
    assert_allclose(ci['b'][2][1], 1.972, rtol=0.01)
    assert_allclose(ci['b'][5][0], 0.95, rtol=0.01)
    assert_allclose(ci['b'][5][1], 2.01, rtol=0.01)

    if verbose:
        captured = capsys.readouterr()
        assert 'Calculating CI for' in captured.out


def test_confidence_pnames(data, pars):
    """Test if pnames works as expected."""
    minimizer = lmfit.Minimizer(residual, pars, fcn_args=(data))
    out = minimizer.leastsq()

    assert_allclose(out.params['a'], 0.1, rtol=0.01)
    assert_allclose(out.params['b'], 2.0, rtol=0.01)

    ci = lmfit.conf_interval(minimizer, out, p_names=['a'])
    assert 'a' in ci
    assert 'b' not in ci


def test_confidence_bounds_reached(data, pars):
    """Check if conf_interval handles bounds correctly"""

    # Should work
    pars['a'].max = 0.2
    minimizer = lmfit.Minimizer(residual, pars, fcn_args=(data))
    out = minimizer.leastsq()
    out.params['a'].stderr = 1
    lmfit.conf_interval(minimizer, out, verbose=True)

    # Should warn (i.e,. limit < para.min)
    pars['b'].max = 2.03
    pars['b'].min = 1.97
    minimizer = lmfit.Minimizer(residual, pars, fcn_args=(data))
    out = minimizer.leastsq()
    out.params['b'].stderr = 0.005
    out.params['a'].stderr = 0.01
    with pytest.warns(UserWarning, match="Bound reached"):
        lmfit.conf_interval(minimizer, out, verbose=True)

    # Should warn (i.e,. limit > para.max)
    out.params['b'].stderr = 0.1
    with pytest.warns(UserWarning, match="Bound reached"):
        lmfit.conf_interval(minimizer, out, verbose=True)


def test_confidence_sigma_vs_prob(data, pars):
    """Calculate confidence by specifying sigma or probability."""
    minimizer = lmfit.Minimizer(residual, pars, fcn_args=(data))
    out = minimizer.leastsq()

    ci_sigma_None = lmfit.conf_interval(minimizer, out, sigmas=None)
    ci_sigmas = lmfit.conf_interval(minimizer, out, sigmas=[1, 2, 3])
    ci_1sigma = lmfit.conf_interval(minimizer, out, sigmas=[1])
    ci_probs = lmfit.conf_interval(minimizer, out,
                                   sigmas=[0.68269, 0.9545, 0.9973])

    assert ci_sigma_None == ci_sigmas
    assert_allclose(ci_sigmas['a'][0][1], ci_probs['a'][0][1], rtol=0.01)
    assert_allclose(ci_sigmas['b'][2][1], ci_probs['b'][2][1], rtol=0.01)
    assert len(ci_1sigma['a']) == 3
    assert len(ci_probs['a']) == 7


def test_confidence_exceptions(data, pars):
    """Make sure the proper exceptions are raised when needed."""
    minimizer = lmfit.Minimizer(residual, pars, calc_covar=False,
                                fcn_args=data)
    out = minimizer.minimize(method='nelder')
    out_lsq = minimizer.minimize(params=out.params, method='leastsq')

    # no uncertainty estimated
    msg = 'Cannot determine Confidence Intervals without sensible uncertainty'
    with pytest.raises(lmfit.MinimizerException, match=msg):
        lmfit.conf_interval(minimizer, out)

    # uncertainty is NaN
    out_lsq.params['a'].stderr = np.nan
    with pytest.raises(lmfit.MinimizerException, match=msg):
        lmfit.conf_interval(minimizer, out_lsq)

    # only one varying parameter
    out_lsq.params['a'].vary = False
    msg = r'Cannot determine Confidence Intervals with < 2 variables'
    with pytest.raises(lmfit.MinimizerException, match=msg):
        lmfit.conf_interval(minimizer, out_lsq)


def test_confidence_warnings(data, pars):
    """Make sure the proper warnings are emitted when needed."""
    minimizer = lmfit.Minimizer(residual, pars, fcn_args=data)
    out = minimizer.minimize(method='leastsq')

    with pytest.warns(UserWarning) as record:
        lmfit.conf_interval(minimizer, out, maxiter=1)
        assert 'maxiter=1 reached and prob' in str(record[0].message)


def test_confidence_with_trace(data, pars):
    """Test calculation of confidence intervals with trace."""
    minimizer = lmfit.Minimizer(residual, pars, fcn_args=data)
    out = minimizer.leastsq()

    ci, tr = lmfit.conf_interval(minimizer, out, sigmas=[0.6827], trace=True)
    for p in out.params:
        diff1 = ci[p][1][1] - ci[p][0][1]
        diff2 = ci[p][2][1] - ci[p][1][1]
        stderr = out.params[p].stderr
        assert abs(diff1 - stderr) / stderr < 0.05
        assert abs(diff2 - stderr) / stderr < 0.05

        assert p in tr.keys()
        assert 'prob' in tr[p].keys()


def test_confidence_2d(data, pars):
    """Test the 2D confidence interval calculation."""
    minimizer = lmfit.Minimizer(residual, pars, fcn_args=data)
    out = minimizer.minimize(method='leastsq')

    cx, cy, grid = lmfit.conf_interval2d(minimizer, out, 'a', 'b', 30, 20)
    assert len(cx.ravel()) == 30
    assert len(cy.ravel()) == 20
    assert grid.shape == (20, 30)


def test_confidence_2d_limits(data, pars):
    """Test the 2D confidence interval calculation using limits."""
    minimizer = lmfit.Minimizer(residual, pars, fcn_args=data)
    out = minimizer.minimize(method='leastsq')

    lim = ((1.0e-6, 0.02), (1.0e-6, -4.0))
    cx, cy, grid = lmfit.conf_interval2d(minimizer, out, 'a', 'b', limits=lim)
    assert grid.shape == (10, 10)
    assert_allclose(min(cx.ravel()), 1.0e-6)
    assert_allclose(max(cx.ravel()), 0.02)
    assert_allclose(min(cy.ravel()), -4.0)
    assert_allclose(max(cy.ravel()), 1.0e-6)


def test_confidence_prob_func(data, pars):
    """Test conf_interval with alternate prob_func."""
    minimizer = lmfit.Minimizer(residual, pars, fcn_args=data)
    out = minimizer.minimize(method='leastsq')
    called = 0

    def my_f_compare(best_fit, new_fit):
        nonlocal called
        called += 1
        nfree = best_fit.nfree
        nfix = best_fit.nvarys - new_fit.nvarys
        dchi = new_fit.chisqr / best_fit.chisqr - 1.0
        return f.cdf(dchi * nfree / nfix, nfix, nfree)

    lmfit.conf_interval(minimizer, out, sigmas=[1], prob_func=my_f_compare)
    assert called > 10



"""Tests for the Parameter class."""

from math import trunc
import pickle

import numpy as np
from numpy.testing import assert_allclose
import pytest

import lmfit


@pytest.fixture
def parameters():
    """Initialize a Parameters class for tests."""
    pars = lmfit.Parameters()
    pars.add(lmfit.Parameter(name='a', value=10.0, vary=True, min=-100.0,
                             max=100.0, expr=None, brute_step=5.0,
                             user_data=1))
    pars.add(lmfit.Parameter(name='b', value=0.0, vary=True, min=-250.0,
                             max=250.0, expr="2.0*a", brute_step=25.0,
                             user_data=2.5))
    exp_attr_values_A = ('a', 10.0, True, -100.0, 100.0, None, 5.0, 1)
    exp_attr_values_B = ('b', 20.0, False, -250.0, 250.0, "2.0*a", 25.0, 2.5)
    assert_parameter_attributes(pars['a'], exp_attr_values_A)
    assert_parameter_attributes(pars['b'], exp_attr_values_B)
    return pars, exp_attr_values_A, exp_attr_values_B


@pytest.fixture
def parameter():
    """Initialize parameter for tests."""
    param = lmfit.Parameter(name='a', value=10.0, vary=True, min=-100.0,
                            max=100.0, expr=None, brute_step=5.0, user_data=1)
    expected_attribute_values = ('a', 10.0, True, -100.0, 100.0, None, 5.0, 1)
    assert_parameter_attributes(param, expected_attribute_values)
    return param, expected_attribute_values


def assert_parameter_attributes(par, expected):
    """Assert that parameter attributes have the expected values."""
    par_attr_values = (par.name, par._val, par.vary, par.min, par.max,
                       par._expr, par.brute_step, par.user_data)
    assert par_attr_values == expected


in_out = [(lmfit.Parameter(name='a'),  # set name
           ('a', -np.inf, True, -np.inf, np.inf, None, None, None)),
          (lmfit.Parameter(name='a', value=10.0),  # set value
           ('a', 10.0, True, -np.inf, np.inf, None, None, None)),
          (lmfit.Parameter(name='a', vary=False),  # fix parameter, set vary to False
           ('a', -np.inf, False, -np.inf, np.inf, None, None, None)),
          (lmfit.Parameter(name='a', min=-10.0),  # set lower bound, value reset to min
           ('a', -10.0, True, -10.0, np.inf, None, None, None)),
          (lmfit.Parameter(name='a', value=-5.0, min=-10.0),  # set lower bound
           ('a', -5.0, True, -10.0, np.inf, None, None, None)),
          (lmfit.Parameter(name='a', max=10.0),  # set upper bound
           ('a', -np.inf, True, -np.inf, 10.0, None, None, None)),
          (lmfit.Parameter(name='a', value=25.0, max=10.0),  # set upper bound, value reset
           ('a', 10.0, True, -np.inf, 10.0, None, None, None)),
          (lmfit.Parameter(name='a', expr="2.0*10.0"),  # set expression, vary becomes False
           ('a', -np.inf, True, -np.inf, np.inf, '2.0*10.0', None, None)),
          (lmfit.Parameter(name='a', brute_step=0.1),  # set brute_step
           ('a', -np.inf, True, -np.inf, np.inf, None, 0.1, None)),
          (lmfit.Parameter(name='a', user_data={'b': {}}),  # set user_data
           ('a', -np.inf, True, -np.inf, np.inf, None, None, {'b': {}}))]


@pytest.mark.parametrize('par, attr_values', in_out)
def test_initialize_Parameter(par, attr_values):
    """Test the initialization of the Parameter class."""
    assert_parameter_attributes(par, attr_values)

    # check for other default attributes
    for attribute in ['_expr', '_expr_ast', '_expr_eval', '_expr_deps',
                      '_delay_asteval', 'stderr', 'correl', 'from_internal',
                      '_val']:
        assert hasattr(par, attribute)


def test_Parameter_no_name():
    """Test for Parameter name, now required positional argument."""
    msg = r"missing 1 required positional argument: 'name'"
    with pytest.raises(TypeError, match=msg):
        lmfit.Parameter()


def test_init_bounds():
    """Tests to make sure that initial bounds are consistent.

    Only for specific cases not tested above with the initializations of the
    Parameter class.

    """
    # test 1: min > max; should swap min and max
    par = lmfit.Parameter(name='a', value=0.0, min=10.0, max=-10.0)
    assert par.min == -10.0
    assert par.max == 10.0

    # test 2: min == max; should raise a ValueError
    msg = r"Parameter 'a' has min == max"
    with pytest.raises(ValueError, match=msg):
        par = lmfit.Parameter(name='a', value=0.0, min=10.0, max=10.0)

    # FIXME: ideally this should be impossible to happen ever....
    # perhaps we should add a  setter method for MIN and MAX as well?
    # test 3: max or min is equal to None
    par.min = None
    par._init_bounds()
    assert par.min == -np.inf

    par.max = None
    par._init_bounds()
    assert par.max == np.inf


def test_parameter_set_value(parameter):
    """Test the Parameter.set() function with value."""
    par, initial_attribute_values = parameter

    par.set(value=None)  # nothing should change
    assert_parameter_attributes(par, initial_attribute_values)

    par.set(value=5.0)
    changed_attribute_values = ('a', 5.0, True, -100.0, 100.0, None, 5.0, 1)
    assert_parameter_attributes(par, changed_attribute_values)

    # check if set value works with new bounds, see issue#636
    par.set(value=500.0, min=400, max=600)
    changed_attribute_values2 = ('a', 500.0, True, 400.0, 600.0, None, 5.0, 1)
    assert_parameter_attributes(par, changed_attribute_values2)


def test_parameter_set_vary(parameter):
    """Test the Parameter.set() function with vary."""
    par, initial_attribute_values = parameter

    par.set(vary=None)  # nothing should change
    assert_parameter_attributes(par, initial_attribute_values)

    par.set(vary=False)
    changed_attribute_values = ('a', 10.0, False, -100.0, 100.0, None, 5.0, 1)
    assert_parameter_attributes(par, changed_attribute_values)


def test_parameter_set_min(parameter):
    """Test the Parameter.set() function with min."""
    par, initial_attribute_values = parameter

    par.set(min=None)  # nothing should change
    assert_parameter_attributes(par, initial_attribute_values)

    par.set(min=-50.0)
    changed_attribute_values = ('a', 10.0, True, -50.0, 100.0, None, 5.0, 1)
    assert_parameter_attributes(par, changed_attribute_values)


def test_parameter_set_max(parameter):
    """Test the Parameter.set() function with max."""
    par, initial_attribute_values = parameter

    par.set(max=None)  # nothing should change
    assert_parameter_attributes(par, initial_attribute_values)

    par.set(max=50.0)
    changed_attribute_values = ('a', 10.0, True, -100.0, 50.0, None, 5.0, 1)
    assert_parameter_attributes(par, changed_attribute_values)


def test_parameter_set_expr(parameter):
    """Test the Parameter.set() function with expr.

    Of note, this only tests for setting/removal of the expression; nothing
    else gets evaluated here... More specific tests that require a Parameters
    class can be found below.

    """
    par, _ = parameter

    par.set(expr='2.0*50.0')  # setting an expression, vary --> False
    changed_attribute_values = ('a', 10.0, False, -100.0, 100.0, '2.0*50.0',
                                5.0, 1)
    assert_parameter_attributes(par, changed_attribute_values)

    par.set(expr=None)  # nothing should change
    assert_parameter_attributes(par, changed_attribute_values)

    par.set(expr='')  # should remove the expression
    changed_attribute_values = ('a', 10.0, False, -100.0, 100.0, None, 5.0, 1)
    assert_parameter_attributes(par, changed_attribute_values)


def test_parameters_set_value_with_expr(parameters):
    """Test the Parameter.set() function with value in presence of expr."""
    pars, _, _ = parameters

    pars['a'].set(value=5.0)
    pars.update_constraints()  # update constraints/expressions
    changed_attr_values_A = ('a', 5.0, True, -100.0, 100.0, None, 5.0, 1)
    changed_attr_values_B = ('b', 10.0, False, -250.0, 250.0, "2.0*a", 25.0, 2.5)
    assert_parameter_attributes(pars['a'], changed_attr_values_A)
    assert_parameter_attributes(pars['b'], changed_attr_values_B)

    # with expression present, setting a value works and will leave vary=False
    pars['b'].set(value=1.0)
    pars.update_constraints()  # update constraints/expressions
    changed_attr_values_A = ('a', 5.0, True, -100.0, 100.0, None, 5.0, 1)
    changed_attr_values_B = ('b', 1.0, False, -250.0, 250.0, None, 25.0, 2.5)
    assert_parameter_attributes(pars['a'], changed_attr_values_A)
    assert_parameter_attributes(pars['b'], changed_attr_values_B)


def test_parameters_set_vary_with_expr(parameters):
    """Test the Parameter.set() function with vary in presence of expr."""
    pars, init_attr_values_A, _ = parameters

    pars['b'].set(vary=True)  # expression should get cleared
    pars.update_constraints()  # update constraints/expressions
    changed_attr_values_B = ('b', 20.0, True, -250.0, 250.0, None, 25.0, 2.5)
    assert_parameter_attributes(pars['a'], init_attr_values_A)
    assert_parameter_attributes(pars['b'], changed_attr_values_B)


def test_parameters_set_expr(parameters):
    """Test the Parameter.set() function with expr."""
    pars, init_attr_values_A, init_attr_values_B = parameters

    pars['b'].set(expr=None)  # nothing should change
    pars.update_constraints()  # update constraints/expressions
    assert_parameter_attributes(pars['a'], init_attr_values_A)
    assert_parameter_attributes(pars['b'], init_attr_values_B)

    pars['b'].set(expr='')  # expression should get cleared, vary still False
    pars.update_constraints()  # update constraints/expressions
    changed_attr_values_B = ('b', 20.0, False, -250.0, 250.0, None, 25.0, 2.5)
    assert_parameter_attributes(pars['a'], init_attr_values_A)
    assert_parameter_attributes(pars['b'], changed_attr_values_B)

    pars['a'].set(expr="b/4.0")  # expression should be set, vary --> False
    pars.update_constraints()
    changed_attr_values_A = ('a', 5.0, False, -100.0, 100.0, "b/4.0", 5.0, 1)
    changed_attr_values_B = ('b', 20.0, False, -250.0, 250.0, None, 25.0, 2.5)
    assert_parameter_attributes(pars['a'], changed_attr_values_A)
    assert_parameter_attributes(pars['b'], changed_attr_values_B)


def test_parameter_set_brute_step(parameter):
    """Test the Parameter.set() function with brute_step."""
    par, initial_attribute_values = parameter

    par.set(brute_step=None)  # nothing should change
    assert_parameter_attributes(par, initial_attribute_values)

    par.set(brute_step=0.0)  # brute_step set to None
    changed_attribute_values = ('a', 10.0, True, -100.0, 100.0, None, None, 1)
    assert_parameter_attributes(par, changed_attribute_values)

    par.set(brute_step=1.0)
    changed_attribute_values = ('a', 10.0, True, -100.0, 100.0, None, 1.0, 1)
    assert_parameter_attributes(par, changed_attribute_values)


def test_getstate(parameter):
    """Test for the __getstate__ method."""
    par, _ = parameter
    assert par.__getstate__() == ('a', 10.0, True, None, -100.0, 100.0, 5.0,
                                  None, None, 10, 1)


def test_setstate(parameter):
    """Test for the __setstate__ method."""
    par, initial_attribute_values = parameter
    state = par.__getstate__()

    par_new = lmfit.Parameter('new')
    attributes_new = ('new', -np.inf, True, -np.inf, np.inf, None, None, None)
    assert_parameter_attributes(par_new, attributes_new)

    par_new.__setstate__(state)
    assert_parameter_attributes(par_new, initial_attribute_values)


def test_parameter_pickle_(parameter):
    """Test that we can pickle a Parameter."""
    par, _ = parameter
    pkl = pickle.dumps(par)
    loaded_par = pickle.loads(pkl)

    assert loaded_par == par


def test_repr():
    """Tests for the __repr__ method."""
    par = lmfit.Parameter(name='test', value=10.0, min=0.0, max=20.0)
    assert par.__repr__() == "<Parameter 'test', value=10.0, bounds=[0.0:20.0]>"

    par = lmfit.Parameter(name='test', value=10.0, vary=False)
    assert par.__repr__() == "<Parameter 'test', value=10.0 (fixed), bounds=[-inf:inf]>"

    par.set(vary=True)
    par.stderr = 0.1
    assert par.__repr__() == "<Parameter 'test', value=10.0 +/- 0.1, bounds=[-inf:inf]>"

    par = lmfit.Parameter(name='test', expr='10.0*2.5')
    assert par.__repr__() == "<Parameter 'test', value=-inf, bounds=[-inf:inf], expr='10.0*2.5'>"

    par = lmfit.Parameter(name='test', brute_step=0.1)
    assert par.__repr__() == "<Parameter 'test', value=-inf, bounds=[-inf:inf], brute_step=0.1>"


def test_setup_bounds_and_scale_gradient_methods():
    """Tests for the setup_bounds and scale_gradient methods.

    Make use of the MINUIT-style transformation to obtain the the Parameter
    values and scaling factor for the gradient.
    See: https://lmfit.github.io/lmfit-py/bounds.html

    """
    # situation 1: no bounds
    par_no_bounds = lmfit.Parameter('no_bounds', value=10.0)
    assert_allclose(par_no_bounds.setup_bounds(), 10.0)
    assert_allclose(par_no_bounds.scale_gradient(par_no_bounds.value), 1.0)

    # situation 2: no bounds, min/max set to None after creating the parameter
    # TODO: ideally this should never happen; perhaps use a setter here
    par_no_bounds = lmfit.Parameter('no_bounds', value=10.0)
    par_no_bounds.min = None
    par_no_bounds.max = None
    assert_allclose(par_no_bounds.setup_bounds(), 10.0)
    assert_allclose(par_no_bounds.scale_gradient(par_no_bounds.value), 1.0)

    # situation 3: upper bound
    par_upper_bound = lmfit.Parameter('upper_bound', value=10.0, max=25.0)
    assert_allclose(par_upper_bound.setup_bounds(), 15.968719422671311)
    assert_allclose(par_upper_bound.scale_gradient(par_upper_bound.value),
                    -0.99503719, rtol=1.e-6)

    # situation 4: lower bound
    par_lower_bound = lmfit.Parameter('upper_bound', value=10.0, min=-25.0)
    assert_allclose(par_lower_bound.setup_bounds(), 35.98610843)
    assert_allclose(par_lower_bound.scale_gradient(par_lower_bound.value),
                    0.995037, rtol=1.e-6)

    # situation 5: both lower and upper bounds
    par_both_bounds = lmfit.Parameter('both_bounds', value=10.0, min=-25.0,
                                      max=25.0)
    assert_allclose(par_both_bounds.setup_bounds(), 0.4115168460674879)
    assert_allclose(par_both_bounds.scale_gradient(par_both_bounds.value),
                    -20.976788, rtol=1.e-6)


def test_value_setter(parameter):
    """Tests for the value setter."""
    par, initial_attribute_values = parameter
    assert_parameter_attributes(par, initial_attribute_values)

    par.value = 200.0  # above maximum
    assert_allclose(par.value, 100.0)

    par.value = -200.0  # below minimum
    assert_allclose(par.value, -100.0)

    del par._expr_eval
    par.value = 10.0
    assert_allclose(par.value, 10.0)
    assert hasattr(par, '_expr_eval')


# Tests for magic methods of the Parameter class
def test__array__(parameter):
    """Test the __array__ magic method."""
    par, _ = parameter
    assert np.asarray(par) == np.asarray(10.0)


def test__str__(parameter):
    """Test the __str__ magic method."""
    par, _ = parameter
    assert str(par) == "<Parameter 'a', value=10.0, bounds=[-100.0:100.0], brute_step=5.0>"


def test__abs__(parameter):
    """Test the __abs__ magic method."""
    par, _ = parameter
    assert_allclose(abs(par), 10.0)
    par.set(value=-10.0)
    assert_allclose(abs(par), 10.0)


def test__neg__(parameter):
    """Test the __neg__ magic method."""
    par, _ = parameter
    assert_allclose(-par, -10.0)
    par.set(value=-10.0)
    assert_allclose(-par, 10.0)


def test__pos__(parameter):
    """Test the __pos__ magic method."""
    par, _ = parameter
    assert_allclose(+par, 10.0)
    par.set(value=-10.0)
    assert_allclose(+par, -10.0)


def test__bool__(parameter):
    """Test the __bool__ magic method."""
    par, _ = parameter
    assert bool(par)


def test__int__(parameter):
    """Test the __int__ magic method."""
    par, _ = parameter
    assert isinstance(int(par), int)
    assert_allclose(int(par), 10)


def test__float__(parameter):
    """Test the __float__ magic method."""
    par, _ = parameter
    par.set(value=5)
    assert isinstance(float(par), float)
    assert_allclose(float(par), 5.0)


def test__trunc__(parameter):
    """Test the __trunc__ magic method."""
    par, _ = parameter
    par.set(value=10.5)
    assert isinstance(trunc(par), int)
    assert_allclose(trunc(par), 10)


def test__add__(parameter):
    """Test the __add__ magic method."""
    par, _ = parameter
    assert_allclose(par + 5.25, 15.25)


def test__sub__(parameter):
    """Test the __sub__ magic method."""
    par, _ = parameter
    assert_allclose(par - 5.25, 4.75)


def test__truediv__(parameter):
    """Test the __truediv__ magic method."""
    par, _ = parameter
    assert_allclose(par / 1.25, 8.0)


def test__floordiv__(parameter):
    """Test the __floordiv__ magic method."""
    par, _ = parameter
    par.set(value=5)
    assert_allclose(par // 2, 2)


def test__divmod__(parameter):
    """Test the __divmod__ magic method."""
    par, _ = parameter
    assert_allclose(divmod(par, 3), (3, 1))


def test__mod__(parameter):
    """Test the __mod__ magic method."""
    par, _ = parameter
    assert_allclose(par % 2, 0)
    assert_allclose(par % 3, 1)


def test__mul__(parameter):
    """Test the __mul__ magic method."""
    par, _ = parameter
    assert_allclose(par * 2.5, 25.0)
    assert_allclose(par * -0.1, -1.0)


def test__pow__(parameter):
    """Test the __pow__ magic method."""
    par, _ = parameter
    assert_allclose(par ** 0.5, 3.16227766)
    assert_allclose(par ** 4, 1e4)


def test__gt__(parameter):
    """Test the __gt__ magic method."""
    par, _ = parameter
    assert par < 11
    assert not par < 10


def test__ge__(parameter):
    """Test the __ge__ magic method."""
    par, _ = parameter
    assert par <= 11
    assert par <= 10
    assert not par <= 9


def test__le__(parameter):
    """Test the __le__ magic method."""
    par, _ = parameter
    assert par >= 9
    assert par >= 10
    assert not par >= 11


def test__lt__(parameter):
    """Test the __lt__ magic method."""
    par, _ = parameter
    assert par > 9
    assert not par > 10


def test__eq__(parameter):
    """Test the __eq__ magic method."""
    par, _ = parameter
    assert par == 10
    assert not par == 9


def test__ne__(parameter):
    """Test the __ne__ magic method."""
    par, _ = parameter
    assert par != 9
    assert not par != 10


def test__radd__(parameter):
    """Test the __radd__ magic method."""
    par, _ = parameter
    assert_allclose(5.25 + par, 15.25)


def test__rtruediv__(parameter):
    """Test the __rtruediv__ magic method."""
    par, _ = parameter
    assert_allclose(1.25 / par, 0.125)


def test__rdivmod__(parameter):
    """Test the __rdivmod__ magic method."""
    par, _ = parameter
    assert_allclose(divmod(3, par), (0, 3))


def test__rfloordiv__(parameter):
    """Test the __rfloordiv__ magic method."""
    par, _ = parameter
    assert_allclose(2 // par, 0)
    assert_allclose(20 // par, 2)


def test__rmod__(parameter):
    """Test the __rmod__ magic method."""
    par, _ = parameter
    assert_allclose(2 % par, 2)
    assert_allclose(25 % par, 5)


def test__rmul__(parameter):
    """Test the __rmul__ magic method."""
    par, _ = parameter
    assert_allclose(2.5 * par, 25.0)
    assert_allclose(-0.1 * par, -1.0)


def test__rpow__(parameter):
    """Test the __rpow__ magic method."""
    par, _ = parameter
    assert_allclose(0.5 ** par, 0.0009765625)
    assert_allclose(4 ** par, 1048576)


def test__rsub__(parameter):
    """Test the __rsub__ magic method."""
    par, _ = parameter
    assert_allclose(5.25 - par, -4.75)



import math
from optparse import OptionParser

from numpy.testing import assert_allclose

from lmfit import Parameters, minimize

from .NISTModels import Models, ReadNistData


def ndig(a, b):
    """Precision for NIST values."""
    return round(-math.log10((abs(abs(a)-abs(b)) + 1.e-15) / abs(b)))


ABAR = ' |----------------+----------------+------------------+-------------------|'


def Compare_NIST_Results(DataSet, myfit, params, NISTdata):
    buff = [' ======================================',
            f' {DataSet}: ',
            ' | Parameter Name |  Value Found   |  Certified Value | # Matching Digits |']
    buff.append(ABAR)

    val_dig_min = 200
    err_dig_min = 200
    fmt = ' | %s | % -.7e | % -.7e   | %2i                |'
    for i in range(NISTdata['nparams']):
        parname = f'b{i+1}'
        par = params[parname]
        thisval = par.value
        certval = NISTdata['cert_values'][i]
        vdig = ndig(thisval, certval)
        pname = (parname + ' value ' + ' '*14)[:14]
        buff.append(fmt % (pname, thisval, certval, vdig))
        val_dig_min = min(val_dig_min, vdig)

        thiserr = par.stderr
        certerr = NISTdata['cert_stderr'][i]
        if thiserr is not None and myfit.errorbars:
            edig = ndig(thiserr, certerr)
            ename = (parname + ' stderr' + ' '*14)[:14]
            buff.append(fmt % (ename, thiserr, certerr, edig))
            err_dig_min = min(err_dig_min, edig)

    buff.append(ABAR)
    sumsq = NISTdata['sum_squares']
    try:
        chi2 = myfit.chisqr
        buff.append(' | Sum of Squares | %.7e  | %.7e    |  %2i               |'
                    % (chi2, sumsq, ndig(chi2, sumsq)))
    except Exception:
        pass
    buff.append(ABAR)
    if not myfit.errorbars:
        buff.append(' |          * * * * COULD NOT ESTIMATE UNCERTAINTIES * * * *              |')
        err_dig_min = 0
    if err_dig_min < 199:
        buff.append(f' Worst agreement: {val_dig_min} digits for value, '
                    f'{err_dig_min} digits for error ')
    else:
        buff.append(f' Worst agreement: {val_dig_min} digits')
    return val_dig_min, '\n'.join(buff)


def NIST_Dataset(DataSet, method='leastsq', start='start2',
                 plot=False, verbose=False):

    NISTdata = ReadNistData(DataSet)
    resid, npar, dimx = Models[DataSet]
    y = NISTdata['y']
    x = NISTdata['x']

    params = Parameters()
    for i in range(npar):
        pname = f'b{i+1}'
        pval1 = NISTdata[start][i]
        params.add(pname, value=pval1)
    try:
        myfit = minimize(resid, params, method=method, args=(x,), kws={'y': y}, nan_policy='raise')
    except ValueError:
        if verbose:
            print("Fit failed... nans?")
        return False
    digs, buff = Compare_NIST_Results(DataSet, myfit, myfit.params, NISTdata)
    if verbose:
        print(buff)

    return digs > 2


def build_usage():
    modelnames = []
    ms = ''
    for d in sorted(Models.keys()):
        ms = ms + f' {d} '
        if len(ms) > 55:
            modelnames.append(ms)
            ms = '    '
    modelnames.append(ms)
    modelnames = '\n'.join(modelnames)

    usage = f"""
 === Test Fit to NIST StRD Models ===

usage:
------
    python fit_NIST.py [options] Model Start

where Start is one of 'start1','start2' or 'cert', for different
starting values, and Model is one of

    {modelnames}

if Model = 'all', all models and starting values will be run.

options:
--------
  -m  name of fitting method.  One of:
          leastsq, nelder, powell, lbfgsb, bfgs,
          tnc, cobyla, slsqp, cg, newton-cg
      leastsq (Levenberg-Marquardt) is the default
"""
    return usage
############################


def run_interactive():
    usage = build_usage()
    parser = OptionParser(usage=usage, prog="fit-NIST.py")

    parser.add_option("-m", "--method", dest="method",
                      metavar='METH',
                      default='leastsq',
                      help="set method name, default = 'leastsq'")

    (opts, args) = parser.parse_args()
    dset = ''
    start = 'start2'
    if len(args) > 0:
        dset = args[0]
    if len(args) > 1:
        start = args[1]

    if dset.lower() == 'all':
        tpass = 0
        tfail = 0
        failures = []
        dsets = sorted(Models.keys())
        for dset in dsets:
            for start in ('start1', 'start2', 'cert'):
                if NIST_Dataset(dset, method=opts.method, start=start,
                                plot=False, verbose=True):
                    tpass += 1
                else:
                    tfail += 1
                    failures.append(f"   {dset} (starting at '{start}')")
        print('--------------------------------------')
        print(f' Fit Method: {opts.method} ')
        print(f' Final Results: {tpass} pass, {tfail} fail.')
        print(' Tests Failed for:\n %s' % '\n '.join(failures))
        print('--------------------------------------')
    elif dset not in Models:
        print(usage)
    else:
        return NIST_Dataset(dset, method=opts.method,
                            start=start, plot=False, verbose=True)


def RunNIST_Model(model):

    dset = ReadNistData(model)
    func, npar, dimx = Models[model]
    rss = (func(dset['cert_values'], x=dset['x'], y=dset['y'])**2).sum()
    tiny_rss = 1.e-16
    print(rss, dset['sum_squares'], tiny_rss)

    if dset['sum_squares'] < tiny_rss:
        assert rss < tiny_rss
    else:
        assert_allclose(rss, dset['sum_squares'])

    out1 = NIST_Dataset(model, start='start1', plot=False, verbose=False)
    out2 = NIST_Dataset(model, start='start2', plot=False, verbose=False)
    assert (out1 or out2)


def test_Bennett5():
    RunNIST_Model('Bennett5')


def test_BoxBOD():
    RunNIST_Model('BoxBOD')


def test_Chwirut1():
    RunNIST_Model('Chwirut1')


def test_Chwirut2():
    RunNIST_Model('Chwirut2')


def test_DanWood():
    RunNIST_Model('DanWood')


def test_ENSO():
    RunNIST_Model('ENSO')


def test_Eckerle4():
    RunNIST_Model('Eckerle4')


def test_Gauss1():
    RunNIST_Model('Gauss1')


def test_Gauss2():
    RunNIST_Model('Gauss2')


def test_Gauss3():
    RunNIST_Model('Gauss3')


def test_Hahn1():
    RunNIST_Model('Hahn1')


def test_Kirby2():
    RunNIST_Model('Kirby2')


def test_Lanczos1():
    RunNIST_Model('Lanczos1')


def test_Lanczos2():
    RunNIST_Model('Lanczos2')


def test_Lanczos3():
    RunNIST_Model('Lanczos3')


def test_MGH09():
    RunNIST_Model('MGH09')


def test_MGH10():
    RunNIST_Model('MGH10')


def test_MGH17():
    RunNIST_Model('MGH17')


def test_Misra1a():
    RunNIST_Model('Misra1a')


def test_Misra1b():
    RunNIST_Model('Misra1b')


def test_Misra1c():
    RunNIST_Model('Misra1c')


def test_Misra1d():
    RunNIST_Model('Misra1d')


def test_Nelson():
    RunNIST_Model('Nelson')


def test_Rat42():
    RunNIST_Model('Rat42')


def test_Rat43():
    RunNIST_Model('Rat43')


def test_Roszman1():
    RunNIST_Model('Roszman1')


def test_Thurber():
    RunNIST_Model('Thurber')


if __name__ == '__main__':
    run_interactive()



"""Tests for the AMPGO global minimization algorithm."""

import numpy as np
from numpy.testing import assert_allclose
import pytest

import lmfit
from lmfit._ampgo import ampgo, tunnel

# correct result for Alpine02 function
global_optimum = [7.91705268, 4.81584232]
fglob = -6.12950


@pytest.mark.parametrize("tabustrategy", ['farthest', 'oldest'])
def test_ampgo_Alpine02(minimizer_Alpine02, tabustrategy):
    """Test AMPGO algorithm on Alpine02 function."""
    kws = {'tabustrategy': tabustrategy}
    out = minimizer_Alpine02.minimize(method='ampgo', **kws)
    out_x = np.array([out.params['x0'].value, out.params['x1'].value])

    assert_allclose(out.residual, fglob, rtol=1e-5)
    assert_allclose(min(out_x), min(global_optimum), rtol=1e-3)
    assert_allclose(max(out_x), max(global_optimum), rtol=1e-3)
    assert 'global' in out.ampgo_msg


def test_ampgo_bounds(minimizer_Alpine02):
    """Test AMPGO algorithm with bounds."""
    # change boundaries of parameters
    pars_bounds = lmfit.Parameters()
    pars_bounds.add_many(('x0', 1., True, 5.0, 15.0),
                         ('x1', 1., True, 2.5, 7.5))

    out = minimizer_Alpine02.minimize(params=pars_bounds, method='ampgo')
    assert 5.0 <= out.params['x0'].value <= 15.0
    assert 2.5 <= out.params['x1'].value <= 7.5


def test_ampgo_disp_true(minimizer_Alpine02, capsys):
    """Test AMPGO algorithm with disp is True."""
    # disp to False for L-BFGS-B to avoid too much output...
    kws = {'disp': True, 'local_opts': {'disp': False}}
    minimizer_Alpine02.minimize(method='ampgo', **kws)
    captured = capsys.readouterr()
    assert "Starting MINIMIZATION Phase" in captured.out


def test_ampgo_maxfunevals(minimizer_Alpine02):
    """Test AMPGO algorithm with maxfunevals."""
    # disp to False for L-BFGS-B to avoid too much output...
    kws = {'maxfunevals': 5, 'disp': True, 'local_opts': {'disp': False}}
    out = minimizer_Alpine02.minimize(method='ampgo', **kws)

    assert out.ampgo_msg == 'Maximum number of function evaluations exceeded'


def test_ampgo_local_solver(minimizer_Alpine02):
    """Test AMPGO algorithm with local solver."""
    kws = {'local': 'Nelder-Mead'}

    out = minimizer_Alpine02.minimize(method='ampgo', **kws)
    out_x = np.array([out.params['x0'].value, out.params['x1'].value])

    assert 'ampgo' and 'Nelder-Mead' in out.method
    assert_allclose(out.residual, fglob, rtol=1e-5)
    assert_allclose(min(out_x), min(global_optimum), rtol=1e-3)
    assert_allclose(max(out_x), max(global_optimum), rtol=1e-3)
    assert 'global' in out.ampgo_msg


def test_ampgo_invalid_local_solver(minimizer_Alpine02):
    """Test AMPGO algorithm with invalid local solvers."""
    kws = {'local': 'leastsq'}
    with pytest.raises(Exception, match=r'Invalid local solver selected'):
        minimizer_Alpine02.minimize(method='ampgo', **kws)


def test_ampgo_invalid_tabulistsize(minimizer_Alpine02):
    """Test AMPGO algorithm with invalid tabulistsize."""
    kws = {'tabulistsize': 0}
    with pytest.raises(Exception, match=r'Invalid tabulistsize specified'):
        minimizer_Alpine02.minimize(method='ampgo', **kws)


def test_ampgo_invalid_tabustrategy(minimizer_Alpine02):
    """Test AMPGO algorithm with invalid tabustrategy."""
    kws = {'tabustrategy': 'unknown'}
    with pytest.raises(Exception, match=r'Invalid tabustrategy specified'):
        minimizer_Alpine02.minimize(method='ampgo', **kws)


def test_ampgo_local_opts(minimizer_Alpine02):
    """Test AMPGO algorithm, pass local_opts to solver."""
    # use local_opts to pass maxfun to the local optimizer: providing a string
    # whereas an integer is required, this should throw an error.
    kws = {'local_opts': {'maxfun': 'string'}}
    with pytest.raises(TypeError):
        minimizer_Alpine02.minimize(method='ampgo', **kws)

    # for coverage: make sure that both occurrences are reached
    kws = {'local_opts': {'maxfun': 10}, 'maxfunevals': 50}
    minimizer_Alpine02.minimize(method='ampgo', **kws)


def test_ampgo_incorrect_length_for_bounds():
    """Test for ValueError when bounds are given for only some parameters."""
    def func(x):
        return x[0]**2 + 10.0*x[1]

    msg = r'length of x0 != length of bounds'
    with pytest.raises(ValueError, match=msg):
        ampgo(func, x0=[0, 1], bounds=[(-10, 10)])


def test_ampgo_tunnel_more_than_three_arguments():
    """Test AMPGO tunnel function with more than three arguments."""
    def func(x, val):
        return x[0]**2 + val*x[1]

    args = [func, 0.1, np.array([1, 2]), 5.0]
    out = tunnel(np.array([10, 5]), *args)
    assert_allclose(out, 185.386275588)



#!/usr/bin/env python
"""
Fitting data with uncertainties in x and y
==============================================

This examples shows a general way of fitting a model
to y(x) data which has uncertainties in both y and x.

For more in-depth discussion, see
    https://dx.doi.org/10.1021/acs.analchem.0c02178
"""

import matplotlib.pyplot as plt
import numpy as np

from lmfit import Minimizer, Parameters, report_fit
from lmfit.lineshapes import gaussian

# create data to be fitted
np.random.seed(17)
xtrue = np.linspace(0, 50, 101)
xstep = xtrue[1] - xtrue[0]
amp, cen, sig, offset, slope = 39, 28.2, 4.4, -13, 0.012

ytrue = (gaussian(xtrue, amplitude=amp, center=cen, sigma=sig)
         + offset + slope * xtrue)

ydat = ytrue + np.random.normal(size=xtrue.size, scale=0.1)

# we add errors to x after y has been created, as if there is
# an ideal y(x) and we have noise in both x and y.
# we force the uncertainty away from 'normal', forcing
# it to be smaller than the step size.
xerr = np.random.normal(size=xtrue.size, scale=0.1*xstep)
max_xerr = 0.8*xstep
xerr[np.where(xerr > max_xerr)] = max_xerr
xerr[np.where(xerr < -max_xerr)] = -max_xerr
xdat = xtrue + xerr

# now we assert that we know the uncertaintits in y and x
#   we'll pick values that are reesonable but not exactly
#   what we used to make the noise
yerr = 0.06
xerr = xstep


def peak_model(params, x):
    """Model a peak with a linear background."""
    amp = params['amp'].value
    cen = params['cen'].value
    sig = params['sig'].value
    offset = params['offset'].value
    slope = params['slope'].value
    return offset + slope * x + gaussian(x, amplitude=amp, center=cen, sigma=sig)


# objective without xerr
def objective_no_xerr(params, x, y, yerr):
    model = peak_model(params, x)
    return (model - y) / abs(yerr)


# objective with xerr
def objective_with_xerr(params, x, y, yerr, xerr):
    model = peak_model(params, x)
    dmodel_dx = np.gradient(model) / np.gradient(x)
    dmodel = np.sqrt(yerr**2 + (xerr*dmodel_dx)**2)
    return (model - y) / dmodel


# create a set of Parameters
params = Parameters()
params.add('amp', value=50, min=0)
params.add('cen', value=25)
params.add('sig', value=10)
params.add('slope', value=1.e-4)
params.add('offset', value=-5)

# first fit without xerr
mini1 = Minimizer(objective_no_xerr, params, fcn_args=(xdat, ydat, yerr))
result1 = mini1.minimize()
bestfit1 = peak_model(result1.params, xdat)


mini2 = Minimizer(objective_with_xerr, params, fcn_args=(xdat, ydat, yerr, xerr))
result2 = mini2.minimize()

bestfit2 = peak_model(result2.params, xdat)


print("### not including uncertainty in x:")
print(report_fit(result1))
print("### including uncertainty in x:")
print(report_fit(result2))

print(xdat[:4])

plt.plot(xdat, ydat, 'o', label='data with noise in x and y')
plt.plot(xtrue, ytrue, '-+', label='true data')
plt.plot(xdat, bestfit1, label='fit, no x error')
plt.plot(xdat, bestfit2, label='fit, with x error')
plt.legend()
plt.show()

# # <end examples/doc_uncertainties_in_x_and_y.py>
